/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2013-2014 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of the GNU General
 * Public License Version 2 only ("GPL"). You may not use this file except
 * in compliance with the License.  You can obtain a copy of the License at
 * https://avatar.java.net/license.html or legal/LICENSE.txt.  See the
 * License for the specific language governing permissions and limitations
 * under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 */

/**
 * Avatar Model-Store API. This API supports the notion of models and
 * stores. A model can be schema-based or schema-less. Stores can be either
 * relational or non-relational. Models can be associated with a particular
 * store via a <em>bind</em> operation.
 *
 * @module org/glassfish/avatar-model
 */

'use strict';

var common = require('org/glassfish/avatar/common');
var Promise = common.Promise;
var wrapCallback = common.wrapCallback;
var StoreHelper = Java.type('org.glassfish.avatar.runtime.service.model.StoreHelper');
var log = common.log;
var when = require('when');
var sequence = require('when/sequence');
var parallel = require('when/parallel');

/**
 * <p>Creates a new model type given a name and (optionally) a schema. The model
 * type returned is a constructor function of which instances can be created.
 * A schema is represented as an object literal and each property must
 * be assigned one of the following data types: <tt>"string"</tt>, <tt>"number"</tt>,
 * <tt>"integer"</tt>, <tt>"decimal"</tt>, <tt>"boolean"</tt>, <tt>"date"</tt>,
 * <tt>"object"</tt> or <tt>"array"</tt>. For example:
 * <pre>
 *     var Product = avatar.newModel('product', {
 *      "name": {
 *           "type": "string",
 *           "primary": true
 *       },
 *       "price": "number",
 *       "quantity": "integer"
 *     });
 * </pre>
 *
 * <p>Zero or more properties can be defined as comprising the primary key;
 * in the case of zero, the system will automatically create a primary
 * named "id_" of type "string".
 *
 * <p>If a schema is not provided, then the model is assumed to be
 * schema-less. Most stores will serialize a schema-less model
 * instance into JSON for storing purposes. Properties of type "object"
 * and "array" may be serialized into JSON as well.
 *
 * <p>All the callback functions in this module are of the form:
 * <pre>
 *    function(err, value) { ... }
 * </pre>
 * and all methods return a promise if no callback is specified.
 *
 * @param {type} name Model type name.
 * @param {type} schema Schema definition for the model type (optional).
 * @returns {Model} A constructor function that can be used to create instances
 *  or directly by calling its methods.
 */
exports.newModel = function(name, schema) {
    var cons = function(obj) {
            if (obj === undefined) {
                obj = {};
            }
            for (var p in schema) {
                this[p] = obj[p];
            }
            this.metadata = cons.metadata;
        };

    var metadata = {
        name: name,
        schema: schema,
        oneToMany: [],
        manyToOne: [],
        oneToOne: [],
        store: null
    };

    cons.metadata = function() {
        return metadata;
    };

    var checkMetadata = function(obj, metadata) {
            if (metadata !== obj.metadata()) {
                throw "Use method create() for objects that were not created using new";
            }
        };

    var prepareModel = function(modelCons, obj) {
            if (obj === null) {
                return;
            }

            // Extend object with meta-data
            var metadata = modelCons.metadata();
            obj.metadata = function() {
                return metadata;
            };

            // Prepare OneToMany relationship
            function prepareOneToMany(rel) {
                var model = rel.model;
                var props = rel.properties;

                // Define "as" property as non-enumerable to avoid its serialization
                Object.defineProperty(obj, props.as, {
                    writable: false,
                    configurable: true,
                    enumerable: false,
                    // skip in JSON.stringify()
                    value: {
                        create: function(child, cb) {
                            var deferred = Promise.defer();
                            cons.store().helper.createChildEntity(obj.metadata().name, obj, model.metadata().name,
                                    props.as, child, modelCons.store().getTx(), function(event, args) {
                                var instance = args[0];
                                prepareModel(model, instance);
                                deferred.resolve(instance);
                            }, function(event, args) {
                                deferred.reject(args.length > 0 ? args[0] : event);
                            });
                            return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
                        },
                        get: function() {
                            // Collect arguments ID and callback
                            var id, cb = undefined;
                            var last = arguments[arguments.length - 1]; // last argument
                            if (typeof last === "function") {
                                cb = last;
                                id = Array.prototype.slice.call(arguments, 0, arguments.length - 1);
                            } else {
                                id = Array.prototype.slice.call(arguments, 0, arguments.length);
                            }
                            var deferred = Promise.defer();
                            cons.store().helper.getChildEntity(obj.metadata().name, obj, model.metadata().name,
                                    props.as, id, modelCons.store().getTx(), function(event, args) {
                                if (args.length > 0) {
                                    var instance = args[0];
                                    prepareModel(cons, instance);
                                    deferred.resolve(instance);
                                } else {
                                    deferred.resolve(null);
                                }
                            }, function(event, args) {
                                deferred.reject(args.length > 0 ? args[0] : event);
                            });
                            return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
                        },
                        put: function(child, cb) {
                            var deferred = Promise.defer();
                            cons.store().helper.putChildEntity(obj.metadata().name, obj, model.metadata().name,
                                    props.as, child, modelCons.store().getTx(), function(event, args) {
                                if (args.length > 0) {
                                    var instance = args[0];
                                    prepareModel(model, instance);
                                    deferred.resolve(instance);
                                } else {
                                    deferred.resolve(null);
                                }
                            }, function(event, args) {
                                deferred.reject(args.length > 0 ? args[0] : event);
                            });
                            return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
                        },
                        delete: function() {
                            // Collect arguments ID and callback
                            var id, cb = undefined;
                            var last = arguments[arguments.length - 1]; // last argument
                            if (typeof last === "function") {
                                cb = last;
                                id = Array.prototype.slice.call(arguments, 0, arguments.length - 1);
                            } else {
                                id = Array.prototype.slice.call(arguments, 0, arguments.length);
                            }
                            var deferred = Promise.defer();
                            cons.store().helper.deleteChildEntity(obj.metadata().name, obj, model.metadata().name,
                                    props.as, id, modelCons.store().getTx(), function(event, args) {
                                deferred.resolve();
                            }, function(event, args) {
                                deferred.reject(args.length > 0 ? args[0] : event);
                            });
                            return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
                        },
                        getAll: function(cb) {
                            var deferred = Promise.defer();
                            cons.store().helper.getAllChildEntities(obj.metadata().name, obj, model.metadata().name,
                                    props.as, modelCons.store().getTx(), function(event, args) {
                                var instances = args[0];
                                instances.forEach(function(instance) {
                                    prepareModel(model, instance);
                                });
                                deferred.resolve(instances);
                            }, function(event, args) {
                                deferred.reject(args.length > 0 ? args[0] : event);
                            });
                            return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
                        },
                        deleteAll: function(cb) {
                            var deferred = Promise.defer();
                            cons.store().helper.deleteAllChildEntities(obj.metadata().name, obj, model.metadata().name,
                                    props.as, modelCons.store().getTx(), function(event, args) {
                                deferred.resolve();
                            }, function(event, args) {
                                deferred.reject(args.length > 0 ? args[0] : event);
                            });
                            return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
                        }
                    }
                });
            }

            // Prepare ManyToOne relationship
            function prepareManyToOne(rel) {
                var model = rel.model;
                var props = rel.properties;

                // Define "as" property as non-enumerable to avoid its serialization
                Object.defineProperty(obj, props.as, {
                    writable: true,
                    configurable: true,
                    enumerable: false,
                    // skip in JSON.stringify()
                    value: {
                        get: function(cb) {
                            var deferred = Promise.defer();
                            cons.store().helper.getRelatedEntity(obj.metadata().name, obj, model.metadata().name,
                                    props.as, modelCons.store().getTx(), function(event, args) {
                                var instance = args[0];
                                prepareModel(model, instance);
                                deferred.resolve(instance);
                            }, function(event, args) {
                                deferred.reject(args.length > 0 ? args[0] : event);
                            });
                            return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
                        },
                        put: function(value, cb) {
                            var deferred = Promise.defer();
                            cons.store().helper.putRelatedEntity(obj.metadata().name, obj, model.metadata().name,
                                    props.as, value, modelCons.store().getTx(), function(event, args) {
                                var instance = args[0];
                                prepareModel(modelCons, instance);
                                deferred.resolve(instance);
                            }, function(event, args) {
                                deferred.reject(args.length > 0 ? args[0] : event);
                            });
                            return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
                        }
                    }
                });
            }

            // Prepare OneToOne relationship
            function prepareOneToOne(rel) {
                var model = rel.model;
                var props = rel.properties;

                prepareManyToOne(rel);

                // Add a create operation for 1:1
                obj[props.as].create = function(cobj, cb) {
                    var deferred = Promise.defer();
                    model.create(cobj).then(function(cobj) {
                        obj[props.as].put(cobj).then(function() {
                            deferred.resolve(cobj);
                        }).otherwise(function() {
                            deferred.reject();
                        });
                    });
                    return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
                };
            }

            metadata.oneToMany.forEach(prepareOneToMany);
            metadata.manyToOne.forEach(prepareManyToOne);
            metadata.oneToOne.forEach(prepareOneToOne);
        };

    // --- Methods -------------------------------------------------------
    cons.create = function(obj, cb) {
        var deferred = Promise.defer();
        if (this.isBound()) { // connected?
            this.store().helper.createEntity(this.metadata().name, obj, this.store().getTx(), function(event, args) {
                var instance = args[0];
                prepareModel(cons, instance);
                deferred.resolve(instance);
            }, function(event, args) {
                deferred.reject(args.length > 0 ? args[0] : event);
            });
        } else {
            deferred.resolve(this);
        }
        return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
    };

    cons.get = function() {
        // Collect arguments ID and callback
        var id, cb = undefined;
        var last = arguments[arguments.length - 1]; // last argument
        if (typeof last === "function") {
            cb = last;
            id = Array.prototype.slice.call(arguments, 0, arguments.length - 1);
        } else {
            id = Array.prototype.slice.call(arguments, 0, arguments.length);
        }

        var deferred = Promise.defer();
        if (this.isBound()) { // connected?
            this.store().helper.getEntity(this.metadata().name, id, this.store().getTx(), function(event, args) {
                if (args.length > 0) {
                    var instance = args[0];
                    prepareModel(cons, instance);
                    deferred.resolve(instance);
                } else {
                    deferred.resolve(null);
                }
            }, function(event, args) {
                deferred.reject(args.length > 0 ? args[0] : event);
            });
        } else {
            return null;
        }
        return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
    };

    cons.put = function(obj, cb) {
        checkMetadata(obj, this.metadata());
        var deferred = Promise.defer();
        if (this.isBound()) { // connected?
            this.store().helper.putEntity(this.metadata().name, obj, this.store().getTx(), function(event, args) {
                var instance = args[0];
                prepareModel(cons, instance);
                deferred.resolve(instance);
            }, function(event, args) {
                deferred.reject(args.length > 0 ? args[0] : event);
            });
        } else {
            return null;
        }
        return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
    };

    cons.delete = function() {
        // Collect arguments ID and callback
        var id, cb = undefined;
        var last = arguments[arguments.length - 1]; // last argument
        if (typeof last === "function") {
            cb = last;
            id = Array.prototype.slice.call(arguments, 0, arguments.length - 1);
        } else {
            id = Array.prototype.slice.call(arguments, 0, arguments.length);
        }

        var deferred = Promise.defer();
        if (this.isBound()) { // connected?
            this.store().helper.deleteEntity(this.metadata().name, id, this.store().getTx(), function(event, args) {
                deferred.resolve();
            }, function(event, args) {
                deferred.reject(args.length > 0 ? args[0] : event);
            });
        } else {
            return null;
        }
        return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
    };

    cons.getAll = function(cb) {
        var deferred = Promise.defer();
        if (this.isBound()) { // connected?
            this.store().helper.getAllEntities(this.metadata().name, this.store().getTx(), function(event, args) {
                if (args.length > 0) {
                    var instances = args[0];
                    instances.forEach(function(instance) {
                        prepareModel(cons, instance);
                    });
                    deferred.resolve(instances);
                } else {
                    deferred.resolve(null);
                }
            }, function(event, args) {
                deferred.reject(args.length > 0 ? args[0] : event);
            });
        } else {
            return null;
        }
        return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
    };

    cons.deleteAll = function(cb) {
        var deferred = Promise.defer();
        if (this.isBound()) { // connected?
            this.store().helper.deleteAllEntities(this.metadata().name, this.store().getTx(), function(event, args) {
                deferred.resolve();
            }, function(event, args) {
                deferred.reject(args.length > 0 ? args[0] : event);
            });
        } else {
            return null;
        }
        return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
    };

    cons.query = function(qry, cb) {
        var deferred = Promise.defer();
        if (this.isBound()) { // connected?
            this.store().helper.queryEntities(this.metadata().name, qry, this.store().getTx(), function(event, args) {
                if (args.length > 0) {
                    var instances = args[0];
                    instances.forEach(function(instance) {
                        prepareModel(cons, instance);
                    });
                    deferred.resolve(instances);
                } else {
                    deferred.resolve(null);
                }
            }, function(event, args) {
                deferred.reject(args.length > 0 ? args[0] : event);
            });
        } else {
            return null;
        }
        return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
    };

    cons.hasMany = function(model, properties) {
        // Define one-to-many relationship
        this.metadata().oneToMany.push({
            model: model,
            modelName: model.metadata().name,
            properties: properties
        });

        // Automatically define reverse many-to-one relationship
        var that = this;
        model.metadata().manyToOne.push({
            model: that,
            modelName: that.metadata().name,
            properties: {
                as: properties.foreign,
                foreign: properties.as
            }
        });
    };

    cons.hasOne = function(model, properties) {
        // Define one-to-one relationship
        this.metadata().oneToOne.push({
            model: model,
            modelName: model.metadata().name,
            properties: properties
        });

        // Automatically define reverse one-to-one relationship
        var that = this;
        model.metadata().oneToOne.push({
            model: that,
            modelName: that.metadata().name,
            properties: {
                as: properties.foreign,
                foreign: properties.as
            }
        });
    };

    cons.isBound = function() {
        return this.store != undefined;
    };

    cons.hasSchema = function() {
        return this.metadata().schema !== undefined;
    };

    cons.bind = function(store) {
        if (this.isBound()) {
            this.store().removeModel(this);
        }
        this.store = function() {
            return store;
        };
        store.addModel(this);
    };

    return cons;
};

var Store = function(engine, properties) {
        var that = this;
        this.engine = engine;
        this.properties = properties;
        this.helper = StoreHelper.newStoreHelper(engine, properties, serviceContainer);
        this.models = [];
        this.txw = undefined;

        this.connect = function(cb) {
            var deferred = Promise.defer();
            this.helper.connect(this.engine, this.properties, that.models.map(function(m) {
                return m.metadata().name;
            }), that.models.map(function(m) {
                return m.metadata();
            }), function() {
                deferred.resolve();
            }, function(event, args) {
                deferred.reject(args.length > 0 ? args[0] : event);
            });
            return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
        };

        this.disconnect = function(cb) {
            var deferred = Promise.defer();
            this.helper.disconnect(function() {
                deferred.resolve();
            }, function(event, args) {
                deferred.reject(args.length > 0 ? args[0] : event);
            });
            return wrapCallback((typeof cb === 'function') && cb, deferred.promise);
        };

        this.transactionSeq = function() {
            var args = Array.prototype.slice.call(arguments, 0, arguments.length);
            return this.transaction(function() {
                return when.all(sequence(args));
            });
        };

        this.transactionPar = function() {
            var args = Array.prototype.slice.call(arguments, 0, arguments.length);
            return this.transaction(function() {
                return when.all(parallel(args));
            });
        };

        this.transaction = function(ftx) {
            var store = this;

            function beginTx() {
                var deferred = Promise.defer();
                store.helper.beginTx(function(event, args) {
                    deferred.resolve(args.length > 0 ? args[0] : event);
                }, function(event, args) {
                    deferred.reject(args.length > 0 ? args[0] : event);
                });
                return deferred.promise;
            };

            function commitTx(tx) {
                var deferred = Promise.defer();
                store.helper.commitTx(tx, function() {
                    deferred.resolve();
                }, function(event, args) {
                    deferred.reject(args.length > 0 ? args[0] : event);
                });
                return deferred.promise;
            };

            function rollbackTx(tx) {
                var deferred = Promise.defer();
                store.helper.rollbackTx(tx, function() {
                    deferred.resolve();
                }, function(event, args) {
                    deferred.reject(args.length > 0 ? args[0] : event);
                });
                return deferred.promise;
            };

            function createTxWrapper() {
                return {
                    onCommitted: function(fc) {
                        this.committedf = fc;
                        return this;
                    },
                    onRolledback: function(fr) {
                        this.rollbackedf = fr;
                        return this;
                    },
                    onCommitting: function(fc) {
                        this.committingf = fc;
                        return this;
                    },
                    onRollingback: function(fr) {
                        this.rollingbackf = fr;
                        return this;
                    },
                    commit: function() {
                        this.committed = true;
                    },
                    rollback: function() {
                        this.rolledback = true;
                    },
                    done: function(tx) {
                        var that = this;
                        if (this.rolledback) {
                            if (this.rollingbackf) {
                                this.rollingbackf.call(null);
                            }
                            rollbackTx(tx).then(function() {
                                store.txw = undefined; // clear from store object
                                if (that.rollbackedf) {
                                    that.rollbackedf.call(null);
                                }
                            });
                        } else if (this.committed) {
                            if (this.committingf) {
                                this.committingf.call(null);
                            }
                            commitTx(tx).then(function() {
                                store.txw = undefined; // clear from store object
                                if (that.committedf) {
                                    that.committedf.call(null);
                                }
                            });
                        }
                    }
                };
            }

            var txw = createTxWrapper();
            beginTx().then(function(tx) {
                txw.tx = tx ? tx : null; // set tx in wrapper
                var promise = ftx.call(null);
                promise.then(function() {
                    txw.commit();
                    txw.done(tx);
                }).otherwise(function() {
                    txw.rollback();
                    txw.done(tx);
                });
            });

            this.txw = txw; // set wrapper in store object
            return txw;
        };

        this.addModel = function(model) {
            this.models.push(model);
        };

        this.removeModel = function(model) {
            this.models = this.models.filter(function(m) {
                return m !== model;
            });
        };

        this.bind = function() {
            var that = this;
            var models = Array.prototype.slice.call(arguments);
            models.forEach(function(m) {
                m.bind(that);
            });
        };

        this.getTx = function() {
            return this.txw ? this.txw.tx : null;
        };
    };

/**
 * Creates a new store given an engine name, unique for each database,
 * and a set of properties. The properties are used to establish a connection
 * and to initialize the store according to the application needs.
 *
 * <p>For example:
 * <pre>
 *     var store = avatar.newStore('mysql', {
 *         host: 'localhost',
 *         port: 3306,
 *         database: 'test',
 *         username: 'root',
 *         password: 'root1',
 *         createDb: true,
 *         dropTables: true,
 *         properties: {
 *             "eclipselink.logging.level": "FINE"
 *         }
 *     });
 *  </pre>
 *  Most properties will be defaulted based on the settings of the database (e.g.,
 *  the default port in MySQL is 3306, so it does not need to be specified). For
 *  those stores that are backed by a JPA (EclipseLink) implementation,
 *  additional properties can be specified as shown above.
 *
 * @param {type} engine Name of database engine.
 * @param {type} properties Connection properties.
 * @returns {Store} Store object that can be used to bind models.
 */
exports.newStore = function(engine, properties) {
    return new Store(engine, properties);
};