/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2013-2014 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of the GNU General
 * Public License Version 2 only ("GPL"). You may not use this file except
 * in compliance with the License.  You can obtain a copy of the License at
 * https://avatar.java.net/license.html or legal/LICENSE.txt.  See the
 * License for the specific language governing permissions and limitations
 * under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 */

/**
 * This module provides threading and scale-out functionality.
 *
 * @module org/glassfish/avatar/threads
 */

'use strict';

var common = require('org/glassfish/avatar/common');
var Promise = common.Promise;
var wrapCallback = common.wrapCallback;
var threadIndex = serviceContainer.getApplicationDescriptor().threadIndex();

var assertValidKey = function (key) {
    if (!key) throw new Error('Invalid key: ' + key);
};

var assertValidValue = function (key, value) {
    if (value === null || typeof value === 'function' || typeof value === 'undefined') {
        throw new Error('Invalid value for ' + key + ': ' + value);
    }
};

var assertFunction = function (callback) {
    if (typeof callback !== 'function') {
        throw new Error("Callback must be a function.")
    }
    return callback;
};

var nullify = function (obj) {
    return obj === undefined ? null : obj;
};

var nullifyFunction = function (func) {
    if (!func) return null;
    return assertFunction(func);
};

/**
 * Executes the supplied function on a separate thread. This function is primarily intended for
 * use with blocking Java methods to avoid long pauses in event handling on the main thread.
 * @param {function} func The function to invoke on a separate thread.
 * @param {function} callback An optional callback function to be called with the result.
 * @returns a promise unless a callback is provided.
 * @static
 */
exports.execute = function (func, callback) {
    var deferred = Promise.defer();
    serviceContainer.execute(func, function (event, args) {
        if (args[0]) {
            deferred.reject(args[0]);
        } else {
            deferred.resolve(args[1]);
        }
    });
    return wrapCallback(callback, deferred.promise);
};

/**
 * Creates a new service request handling thread.
 * @param {string} scriptName An optional script to execute instead of the default service main script.
 * @returns {Thread}
 */
exports.newServiceThread = function (scriptName) {
    var name = serviceContainer.getApplicationDescriptor().name();
    var script = scriptName ? script : null;
    return new Thread(name, true, script);
};

/**
 * Creates a new background thread.
 * @param {string} name The thread name.
 * @param {string} scriptName The name of the script to execute.
 * @returns {Thread}
 */
exports.newBackgroundThread = function (name, scriptName) {
    return new Thread(name, false, scriptName);
};

// ----------- Thread ------------------

var ContainerCallback = Java.type('org.glassfish.avatar.runtime.service.ContainerCallback');

/**
 * A callback function for thread lifecycle events.
 * @callback threadLifecycle
 * @param {string} state The state change: "started" or "stopped".
 */

/**
 * Create a new event driven thread.
 * @param {string} name The thread name (to which an index will be appended).
 * @param {boolean} acceptRequests true if this thread should accept external service requests
 * (so acts like default service threads), or false if not.
 * @param {string} scriptName The name of the script to execute. If not provided, the default
 * service main file name will be used.
 * @constructor
 */
var Thread = function (name, acceptRequests, scriptName) {
    var script = (scriptName === null || scriptName === undefined) ? null : scriptName;
    var requests = (acceptRequests === undefined) ? (script === null) : acceptRequests;
    this.container = servicePool.newContainer(name, requests, script);
};

/**
 * Start the thread.
 * @param {threadLifecycle} callback An optional callback to receive "started" or "stopped" lifecycle
 * events.
 */
Thread.prototype.start = function (callback) {
    this.container.start(callback == null ? null : new ContainerCallback(this.container, function (event, args) {
        callback(args[0]);
    }));
};

/**
 * Block until new thread is started &mdash; should not be used from event threads, pass a
 * callback to the start function instead.
 */
Thread.prototype.awaitStarted = function () {
    this.container.awaitStarted();
};

/**
 * Stop the thread. May not be restarted.
 */
Thread.prototype.stop = function () {
    this.container.shutdown();
    this.container = null;
};

exports.Thread = Thread;


// ----------- SharedState ------------------

var Class = Java.type('java.lang.Class');
var sessions = servicePool.getSessionState();

/**
 * Constructor. An application scoped instance is available from
 * {@link module:org/glassfish/avatar/common.application application.state}.
 *
 * @classdesc A shared key/value store. Enables efficient shared state across threads in
 * the same process and, in clustered runtimes, across hosts.
 * </p>
 * JavaScript values are serialized to preserve thread-safety, and instances are cached for performance.
 * Since serialization of JavaScript values will ignore Java object properties, values are serialized
 * even on the thread that does the put() to ensure that all threads see the same properties.
 * </p>
 * Please note that this copy behavior means that changes to a value will <em>not</em> be seen by other
 * threads unless a put() operation is performed.
 * </p>
 * Java values must be both thread-safe and serializable. The semantics here are different than that for
 * JavaScript values, with priority given to literal sharing instances across threads &mdash; serialization
 * occurs only if the object is transported to another process. This asymmetrical behavior is, in theory,
 * an acceptable trade-off since 1) Java serialization is very expensive, and 2) semantic differences
 * between an original and a deserialized Java object are less likely than with a JavaScript object.
 * <p/>
 * SharedState is an {@link http://nodejs.org/api/events.html#events_class_events_eventemitter EventEmitter}.
 * @constructor
 * @fires added When a value is added -- arguments[0] is the key.
 * @fires added When a value is modified -- arguments[0] is the key.
 * @fires added When a value is removed -- arguments[0] is the key.
 */
var SharedState = common.inheritEventSource(function (name) {
    this.state = sessions.get(name);
    common.EventSource.call(this, this.state);
});

/**
 * Adds a value for the specified key, replacing any existing value.
 * @param {string} key The key.
 * @param value The value.
 * @fires added
 * @fires modified
 */
SharedState.prototype.put = function (key, value) {
    assertValidKey(key);
    assertValidValue(key, value);
    var isJava = value.class instanceof Class;
    this.state.put(threadIndex, key, value, isJava);
};

/**
 * Returns a value for the specified key.
 * @param {string} key The key.
 * @returns value The value, if present.
 */
SharedState.prototype.get = function (key) {
    assertValidKey(key);
    return this.state.get(threadIndex, key);
};

/**
 * Returns true if this map contains the specified value.
 * @param {string} key The key.
 */
SharedState.prototype.contains = function (key) {
    assertValidKey(key);
    return this.state.contains(key);
};

/**
 * Returns all keys in the map.
 * @returns {array} The keys.
 */
SharedState.prototype.keys = function () {
    return this.state.keys();
};

/**
 * Removes the value for the specified key, if present.
 * @param {string} key The key.
 * @fires removed
 */
SharedState.prototype.remove = function (key) {
    assertValidKey(key);
    this.state.remove(threadIndex, key);
};

/**
 * Returns a simple string representation.
 * @returns {string} The string.
 */
SharedState.prototype.toString = function() {
    return this.state.toString();
};

/**
 * Constructs a new session scoped state accessor. The singleton instance is available from
 * {@link module:org/glassfish/avatar/common.application application.sessionState}.
 *
 * @classdesc Provides access to session scoped {@link SharedState state} instances, by name. Session state instances
 * are created on demand, and <b>expire</b> if not accessed within a time defined by 'service.session.ttl.seconds' in
 * avatar.properties. No expired instances will be returned; instead, for the same name, a new instance will be created
 * and returned.
 * <p/>
 * <b>Calls against an expired instance will throw an exception; therefore, best practice is to always request
 * the instance from this sessionState object.</b>
 * <p/>
 * SessionState is an {@link http://nodejs.org/api/events.html#events_class_events_eventemitter EventEmitter}.
 * @constructor
 * @singleton
 * @extends EventSource
 * @fires created When an instance is created -- arguments[0] is the session name.
 * @fires destroyed When an instance expires and is destroyed -- arguments[0] is the session name.
 */
var SessionState = common.inheritEventSource(function () {
    common.EventSource.call(this, sessions);
    this.cache = {};

    this.on('destroyed', function (name) {
        this.cache[name] = null;
    });
});

/**
 * Returns the session scoped state instance for the specified name.
 * @param {string} name The name.
 * @returns {SharedState} The state instance.
 */
SessionState.prototype.get = function (name) {
    var result = this.cache[name];
    if (!result) {
        result = new SharedState(name);
        this.cache[name] = result;
    }
    return result;
};

var applicationState, sessionState; // lazy

exports.applicationState = function () {
    if (!applicationState) {
        applicationState = new SharedState(null);
    }
    return applicationState;
};

exports.sessionState = function () {
    if (!sessionState) {
        sessionState = new SessionState();
    }
    return sessionState;
};

// ----------- MessageBus ------------------

var Bus = Java.type('org.glassfish.avatar.message.MessageBus');
var Channel = Java.type('org.glassfish.avatar.runtime.service.ServiceChannel');

var stringify = function (body) {
    if (!body) throw new Error("Undefined body");
    return JSON.stringify(body);
};

/**
 * A callback function passed to <a href="avatar/msgbus-MessageBus.html">MessageBus</a> methods.
 * @callback messageCallback
 * @param {object} body The deserialized message body.
 * @param {Message} message The message.
 */

/**
 * The message type passed to {@link messageCallback} functions and
 * to the <a href="avatar/msgbus-MessageBus.html#reply">MessageBus.reply()</a> method. Instances
 * are read-only.
 * @typedef Message
 * @property {string} topic The message topic.
 * @property {string} body The serialized message body.
 * @property {string} sourceAddress The address of the source router.
 * @property {string} destinationAddress The address of the destination router, if supplied.
 * @property {string} replyTopic The reply topic, if supplied.
 * @property {boolean} sendAll True if this message was sent to all subscribers, false if only one.
 */

/**
 * Constructor. A singleton instance is available from
 * {@link module:org/glassfish/avatar/common.application application.bus}.
 *
 * @classdesc A lightweight (non-durable) publish-subscribe message bus.
 *
 * The default implementation enables communication between application threads
 * in the local process. Other implementations may support communication between
 * application threads across all cluster members</b>.
 *
 * @example <caption>Publish a simple object to the 'echo' topic.</caption>
 *
 * var avatar = require('org/glassfish/avatar');
 * var bus = avatar.application.bus;
 * bus.publish('echo', { x : 'x', y : 'y' });
 *
 * @example <caption>Subscribe to the 'echo' topic and log all messages. Uses
 * the 'on' alias for the 'subscribe' function.</caption>
 *
 * var avatar = require('org/glassfish/avatar');
 * var bus = avatar.application.bus;
 *
 * bus.on('echo', function(body, msg) {
 *     avatar.log('Got message: ' + JSON.stringify(body));
 * });
 *
 * @example <caption>Subscribe to the 'echo' topic and reply to one message.
 * Reply with the the received body extended with application information,
 * then unsubscribe.</caption>
 *
 * var avatar = require('org/glassfish/avatar');
 * var app = avatar.application;
 * var bus = app.bus;
 *
 * var echoOnce = function(body, msg) {
 *     body.name = app.name;
 *     body.threadIndex = app.threadIndex;
 *     body.threadCount = app.threadCount;
 *     body.startTime = app.startTime;
 *     body.uptime = app.uptime;
 *     body.topic = msg.topic;
 *     body.src = msg.sourceAddress;
 *     bus.reply(msg, body);
 *     bus.unsubscribe(msg.topic, echoOnce);
 * }
 *
 * bus.on('echo', echoOnce);
 * @constructor
 */
var MessageBus = function () {
    this.bus = new Bus(serviceContainer.getPool().getMessageRouter());
};

/**
 * Subscribes to the specified topic.
 * @param {string} topic The topic.
 * @param {messageCallback} callback The callback function to subscribe.
 */
MessageBus.prototype.subscribe = function (topic, callback) {
    assertFunction(callback);
    this.bus.subscribe(topic, new Channel(callback, serviceContainer, function (topic, args) {
        if (args && args.length > 0) {
            var msg = args[0];
            var body = JSON.parse(msg.body);
            callback(body, msg);
        }
    }));
};

/**
 * Subscribes to the specified topic.
 * @param {string} topic The topic.
 * @param {messageCallback} callback The callback function to subscribe.
 */
MessageBus.prototype.on = MessageBus.prototype.subscribe;

/**
 * Unsubscribes a callback from the specified topic.
 * @param {string} topic The topic.
 * @param {messageCallback} callback The callback function to unsubscribe.
 */
MessageBus.prototype.unsubscribe = function (topic, callback) {
    assertFunction(callback);
    this.bus.unsubscribe(topic, new Channel(callback, null, null));
};

/**
 * Send a message to one subscriber of the specified topic. A simple
 * round-robin algorithm is used to select a subscriber.
 * @param {string} topic The topic.
 * @param {object} body The message body, which will be converted to JSON.
 * @param {messageCallback} reply An optional reply callback. If provided,
 * any recipient of the message can send a reply.
 * @returns {boolean} true if there were any subscribers to the topic.
 */
MessageBus.prototype.send = function (topic, body, reply) {
    return this.bus.send(topic, stringify(body), nullifyFunction(reply));
};

/**
 * Send a message to all subscribers of the specified topic, at the specified
 * router address.
 * @param {string} address The router address.
 * @param {string} topic The topic.
 * @param {object} body The message body, which will be converted to JSON.
 * @param {messageCallback} reply An optional reply callback. If provided,
 * any recipient of the message can send a reply.
 * @returns {boolean} true if there were any subscribers to the topic.
 */
MessageBus.prototype.sendTo = function (address, topic, body, reply) {
    return this.bus.sendTo(address, topic, stringify(body), nullifyFunction(reply));
};

/**
 * Send a message to all subscribers of the specified topic.
 * @param {string} topic The topic.
 * @param {object} body The message body, which will be converted to JSON.
 * @param {messageCallback} reply An optional reply callback. If provided,
 * any recipient of the message can send a reply.
 * @returns {boolean} true if there were any subscribers to the topic.
 */
MessageBus.prototype.publish = function (topic, body, reply) {
    return this.bus.publish(topic, stringify(body), nullifyFunction(reply));
};

/**
 * Send a message to all subscribers of the specified topic, at the specified
 * router address.
 * @param {string} address The router address.
 * @param {string} topic The topic.
 * @param {object} body The message body, which will be converted to JSON.
 * @param {messageCallback} reply An optional reply callback. If provided,
 * any recipient of the message can send a reply.
 * @returns {boolean} true if there were any subscribers to the topic.
 */
MessageBus.prototype.publishTo = function (address, topic, body, reply) {
    return this.bus.publishTo(topic, stringify(body), nullifyFunction(reply));
};

/**
 * Reply to a message.
 * @param {Message} replyTo The message to reply to.
 * @param {object} body The message body, which will be converted to JSON.
 * @param {messageCallback} reply An optional reply callback. If provided,
 * any recipient of the message can send a reply.
 * @returns {boolean} true if the replyTo contains valid reply coordinates.
 */
MessageBus.prototype.reply = function (replyTo, body, reply) {
    return this.bus.reply(nullify(replyTo), stringify(body), nullifyFunction(reply));
};

var messageBus; // lazy

exports.messageBus = function () {
    if (!messageBus) {
        messageBus = new MessageBus();
    }
    return messageBus;
};

// ----------- Cluster ------------------

var emptyArray = [];
/**
 * A cluster membership accessor. The singleton instance is available from
 * {@link module:org/glassfish/avatar/common.application application.cluster}.
 * <p/>
 * Cluster is an {@link http://nodejs.org/api/events.html#events_class_events_eventemitter EventEmitter}.
 * @constructor
 * @singleton
 * @extends EventSource
 * @fires added When a member is added -- arguments[0] is the new Member.
 * @fires removed When a member is removed -- arguments[0] is the removed Member.
 */
var Cluster = common.inheritEventSource(function () {
    var self = this;
    this.cluster = servicePool.getCluster();

    common.EventSource.call(this, this.cluster.eventSource(), function (event, args) {
        self.emit(event, new Member(args[0]));
    });

    /**
     * A member descriptor.
     * @property {string} address The member address.
     * @property {number} threadCount The number of threads.
     * @property {Date} startTime The time the member started.
     */
    var Member = function (member) {
        this.address = member.address();
        this.threadCount = member.threadCount();
        this.startTime = new Date(member.startTime());
    };

    this.local = new Member(cluster.localMember());
});


/**
 * Returns the cluster name (i.e. the application name).
 * @returns the name.
 */
Cluster.prototype.name = function () {
    return this.cluster.name();
};

/**
 * Returns the local member.
 * @returns the local member.
 */
Cluster.prototype.localMember = function () {
    return this.local;
};

/**
 * Returns the current list of remote members.
 * @returns {array} The list. Empty if clustering is not enabled or available.
 */
Cluster.prototype.remoteMembers = function () {
    var members = this.cluster.remoteMembers();
    if (members.length == 0) return emptyArray;
    var result = [];
    for (var i = 0, len = members.length; i < len; i++) {
        result[i] = new Member(members[i]);
    }
    return result;
};

var cluster; // lazy

exports.cluster = function () {
    if (!cluster) {
        cluster = new Cluster();
    }
    return cluster;
};

// ----------- Script loading ------------------

/**
 * Loads a script.
 * @param {string} path The script path. Can be a relative path inside the archive, or a relative
 * or absolute path to a file in the file system.
 * @static
 */
exports.load = function (path) {
    return serviceContainer.load(path);
};

/**
 * Loads a script in a new global.
 * @param {string} path The script path. Can be a relative path inside the archive, or a relative
 * or absolute path to a file in the file system.
 * @param {object[]} args The (optional) arguments to the script.
 * @static
 */
exports.loadWithNewGlobal = function (path, args) {
    return serviceContainer.loadWithNewGlobal(path, args);
};
